﻿using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Linq;
using System.Text;
using System.Threading.Tasks;
using System.Windows.Forms;
using System.Diagnostics;
using System.IO;
using Microsoft.CodeAnalysis;
using Microsoft.CodeAnalysis.CSharp;
using Microsoft.CodeAnalysis.CSharp.Syntax;
using System.Text.RegularExpressions;
using System.Reflection;

namespace Converter
{
    public partial class Form1 : Form
    {
        private Nt8 _nt8 = new Nt8();
        public Form1()
        {
            InitializeComponent();
        }

        private void OpenFile_Click(object sender, EventArgs e)
        {
            OpenFileDialog openFileDialog1 = new OpenFileDialog();
            openFileDialog1.DefaultExt = "*.cs";
            openFileDialog1.Filter = "cs files|*.cs";
            openFileDialog1.RestoreDirectory = true;
        
            if (openFileDialog1.ShowDialog() == DialogResult.OK)
            {
                try
                {
                    StreamReader sr = new StreamReader(openFileDialog1.FileName);
                    richTextBox1.Text = sr.ReadToEnd();
                    sr.Close();
                }
                catch (Exception ex)
                {
                    MessageBox.Show("Error: Could not read file from disk. Original error: " + ex.Message);
                }
            }
        }

        private void StartConversion_Click(object sender, EventArgs e)
        {
            richTextBox2.Text = _nt8.ConvertClassText(richTextBox1.Text);
        }

        private void SaveFile_Click(object sender, EventArgs e)
        {
            SaveFileDialog saveFileDialog1 = new SaveFileDialog();

            saveFileDialog1.DefaultExt = "*.cs";
            saveFileDialog1.Filter = "cs files|*.cs";
            saveFileDialog1.RestoreDirectory = true;

            if (saveFileDialog1.ShowDialog() == System.Windows.Forms.DialogResult.OK && saveFileDialog1.FileName.Length > 0)
            {
                try
                {
                    StreamWriter sw = new StreamWriter(saveFileDialog1.FileName);
                    sw.WriteLine(richTextBox2.Text);
                    sw.Close();
                }
                catch (Exception ex)
                {
                    MessageBox.Show("Error: Could not write file from disk. Original error: " + ex.Message);
                }
            }
        }
        public class Nt8
        {
            // Define other methods and classes here
            // List of usings to use for the new class.
            private string _usingsText = @"using System;
                            using System.Collections;
                            using System.Collections.Generic;
                            using System.ComponentModel;
                            using System.ComponentModel.DataAnnotations;
                            using System.Diagnostics.CodeAnalysis;
                            using System.Windows;
                            using System.Reflection;
                            using System.Windows.Media;
                            using System.Xml.Serialization;
                            using NinjaTrader.Cbi;
                            using NinjaTrader.Data;
                            using NinjaTrader.Gui;
                            using NinjaTrader.Gui.Chart;
                            using NinjaTrader.Gui.Tools;
                            using NinjaTrader.NinjaScript.DrawingTools;
                            using NinjaTrader.NinjaScript.Indicators;
                            ";

            // Lookup of types that were renamed.
            // The key is the name of the old type.
            // The value is the name of the new type.
            private Dictionary<string, string> _renamedTypes =
                new Dictionary<string, string>
                {
                    {"System.Drawing.Color", "Brush"},
                    {"Drawing.Color", "Brush"},
                    {"Color", "Brush"},
                    {"SolidBrush", "SolidColorBrush"},
                    {"DashStyle", "DashStyleHelper"},
                    {"StringAlignment", "TextAlignment"},
                    {"IDataSeries", "ISeries<double>"},
                    {"Data.DataSeries", "Series<double>"},
                    {"DataSeries", "Series<double>"},
                    {"IntSeries", "Series<int>"},
                    {"LongSeries", "Series<long>"},
                    {"BoolSeries", "Series<bool>"},
                    {"StringSeries", "Series<string>"},
                    {"TimeSeries", "Series<DateTime>"},
                    {"DateTimeSeries", "Series<DateTime>"},
                    {"ColorSeries", "Series<Brush>"},
                    {"Pen", "Stroke"},
                    {"Font", "SimpleFont"},
                    {"PeriodType", "BarsPeriodType"},
                    {"IOrder", "Order"},
                    {"IExecution", "Execution"},
                    {"IDrawObject", "IDrawingTool"},
                    {"IPosition", "Position"},
                    {"Indicator.Indicator", "Indicators.Indicator"},

                    {"IArc", "Arc"},
                    {"IAndrewsPitchfork", "AndrewsPitchfork"},
                    {"IArrowDown", "ArrowDown"},
                    {"IArrowUp", "ArrowUp"},
                    {"IArrowLine", "ArrowLine"},
                    {"IDiamond", "Diamond"},
                    {"IDot", "Dot"},
                    {"IEllipse", "Ellipse"},
                    {"IExtendedLine", "ExtendedLine"},
                    {"IFibonacciCircle", "FibonacciCircle"},
                    {"IFibonacciExtensions", "FibonacciExtensions"},
                    {"IFibonacciRetracements", "FibonacciRetracements"},
                    {"IFibonacciTimeExtensions", "FibonacciTimeExtensions"},
                    {"IGannFan", "GannFan"},
                    {"IHorizontalLine", "HorizontalLine"},
                    {"ILine", "NinjaTrader.NinjaScript.DrawingTools.Line"},
                    {"IRay", "Ray"},
                    {"IRectangle", "Rectangle"},
                    {"IRegion", "Region"},
                    {"IRegressionChannel", "RegressionChannel"},
                    {"IRuler", "Ruler"},
                    {"ISquare", "Square"},
                    {"IText", "Text"},
                    {"ITextFixed", "TextFixed"},
                    {"ITrendChannel", "TrendChannel"},
                    {"ITriangle", "Triangle"},
                    {"ITriangleDown", "TriangleDown"},
                    {"ITriangleUp", "TriangleUp"},
                    {"IVerticalLine", "VerticalLine"},

                    {"CandleStickPattern", "CandlestickPattern"},
                };

            // Lookup of constructors that were renamed.
            // The key is the name of the old constructor.
            // The value is the name of the new constructor.
            private Dictionary<string, string> _renamedConstructors =
                new Dictionary<string, string>
                {
                    {"IDataSeries", "Series<double>"},
                    {"DataSeries", "Series<double>"},
                    {"IntSeries", "Series<int>"},
                    {"LongSeries", "Series<long>"},
                    {"BoolSeries", "Series<bool>"},
                    {"StringSeries", "Series<string>"},
                    {"TimeSeries", "Series<DateTime>"},
                    {"DateTimeSeries", "Series<DateTime>"},
                    {"ColorSeries", "Series<Brush>"},
                    {"Pen", "Stroke"},
                    {"Font", "SimpleFont"},
                };

            // Lookup of members that were renamed.
            // The key is the name of the old member.
            // The value is the name of the new member.
            private Dictionary<string, string> _renamedMembers =
                new Dictionary<string, string>
                {
                    {"System.Drawing.Color", "Brushes"},
                    {"Drawing.Color", "Brushes"},
                    {"Color", "Brushes"},
                    {"KnownColor", "Brushes"},
                    {"DashStyle", "DashStyleHelper"},
                    {"PeriodType", "BarsPeriodType"},
                    {"StringAlignment", "TextAlignment"},
                    {"StringAlignment.Near", "TextAlignment.Left"},
                    {"StringAlignment.Far", "TextAlignment.Right"},

                    {"CandleStickPattern", "CandlestickPattern"},
                    {"Rising", "IsRising"},
                    {"Falling", "IsFalling"},
                    {"OrderType.Stop", "OrderType.StopMarket"},
                    {"InstrumentType.Currency", "InstrumentType.Forex"},
                    {"Position.AvgPrice", "Position.AveragePrice"},
                    {"Position.GetProfitLoss", "Position.GetUnrealizedProfitLoss"},
                    {"Disable", "SetState(State.Terminated)"},
                    {"GetAccountValue", "Account.Get"},
                    {"Bars.FirstBarOfSession", "Bars.IsFirstBarOfSession"},
                    {"RealtimeErrorHandling.TakeNoAction", "RealtimeErrorHandling.IgnoreAllErrors"},
                    {
                        "RealtimeErrorHandling.StopStrategyCancelOrdersClosePositions",
                        "RealtimeErrorHandling.StopCancelClose"
                    },

                    {"ExitLongStop", "ExitLongStopMarket"},
                    {"ExitShortStop", "ExitShortStopMarket"},
                    {"SubmitOrder", "SubmitOrderUnmanaged"},

                    {"ResetAlert", "RearmAlert"},
                    {"IsValidPlot", "IsValidDataPoint"},

                    {
                        "Performance.AllTrades.TradesPerformance.AvgBarsInTrade",
                        "SystemPerformance.AllTrades.TradesPerformance.AverageBarsInTrade"
                    },
                    {
                        "Performance.AllTrades.TradesPerformance.MaxConsecLoser",
                        "SystemPerformance.AllTrades.TradesPerformance.MaxConsecutiveLoser"
                    },
                    {
                        "Performance.AllTrades.TradesPerformance.MaxConsecWinner",
                        "SystemPerformance.AllTrades.TradesPerformance.MaxConsecutiveWinner"
                    },
                    {
                        "Performance.AllTrades.TradesPerformance.MaxTime2Recover",
                        "SystemPerformance.AllTrades.TradesPerformance.MaxTimeToRecover"
                    },
                    {
                        "Performance.AllTrades.TradesPerformance.Commission",
                        "SystemPerformance.AllTrades.TradesPerformance.TotalCommission"
                    },
                    {
                        "Performance.AllTrades.TradesPerformance.Currency.AvgEtd",
                        "SystemPerformance.AllTrades.TradesPerformance.Currency.AverageEtd"
                    },
                    {
                        "Performance.AllTrades.TradesPerformance.Currency.AvgMae",
                        "SystemPerformance.AllTrades.TradesPerformance.Currency.AverageMae"
                    },
                    {
                        "Performance.AllTrades.TradesPerformance.Currency.AvgMfe",
                        "SystemPerformance.AllTrades.TradesPerformance.Currency.AverageMfe"
                    },
                    {
                        "Performance.AllTrades.TradesPerformance.Currency.AvgProfit",
                        "SystemPerformance.AllTrades.TradesPerformance.Currency.AverageProfit"
                    },
                    {"Performance.RealtimeTrades.Count", "SystemPerformance.RealTimeTrades.Count"},

                    {"NinjaTrader.Gui.Design.SerializableColor.ToString", "Serialize.BrushToString"},
                    {"Gui.Design.SerializableColor.ToString", "Serialize.BrushToString"},
                    {"Design.SerializableColor.ToString", "Serialize.BrushToString"},
                    {"SerializableColor.ToString", "Serialize.BrushToString"},
                    {"NinjaTrader.Gui.Design.SerializableColor.FromString", "Serialize.StringToBrush"},
                    {"Gui.Design.SerializableColor.FromString", "Serialize.StringToBrush"},
                    {"Design.SerializableColor.FromString", "Serialize.StringToBrush"},
                    {"SerializableColor.FromString", "Serialize.StringToBrush"},
                    {"ChartControl.ShowBarsRequired", "/*NT8 Removed: ChartControl.ShowBarsRequired*/false"},
                    {"ChartControl.GetXByBarIdx", "ChartControl.GetXByBarIndex"},
                    {"ChartControl.GetYByValue", "/*chartScale.GetYByValue*/ChartControl.GetYByValue"},
                    {"Instrument.MasterInstrument.Round2TickSize", "Instrument.MasterInstrument.RoundToTickSize"},

                    {"KnownColor.ActiveBorder", "SystemColors.ActiveBorderBrush"},
                    {"KnownColor.ActiveCaption", "SystemColors.ActiveCaptionBrush"},
                    {"KnownColor.ActiveCaptionText", "SystemColors.ActiveCaptionTextBrush"},
                    {"KnownColor.AppWorkspace", "SystemColors.AppWorkspaceBrush"},
                    {"KnownColor.Control", "SystemColors.ControlBrush"},
                    {"KnownColor.ControlDark", "SystemColors.ControlDarkBrush"},
                    {"KnownColor.ControlDarkDark", "SystemColors.ControlDarkDarkBrush"},
                    {"KnownColor.ControlLight", "SystemColors.ControlLightBrush"},
                    {"KnownColor.ControlLightLight", "SystemColors.ControlLightLightBrush"},
                    {"KnownColor.ControlText", "SystemColors.ControlTextBrush"},
                    {"KnownColor.Desktop", "SystemColors.DesktopBrush"},
                    {"KnownColor.GrayText", "SystemColors.GrayTextBrush"},
                    {"KnownColor.Highlight", "SystemColors.HighlightBrush"},
                    {"KnownColor.HighlightText", "SystemColors.HighlightTextBrush"},
                    {"KnownColor.HotTrack", "SystemColors.HotTrackBrush"},
                    {"KnownColor.InactiveBorder", "SystemColors.InactiveBorderBrush"},
                    {"KnownColor.InactiveCaption", "SystemColors.InactiveCaptionBrush"},
                    {"KnownColor.InactiveCaptionText", "SystemColors.InactiveCaptionTextBrush"},
                    {"KnownColor.Info", "SystemColors.InfoBrush"},
                    {"KnownColor.InfoText", "SystemColors.InfoTextBrush"},
                    {"KnownColor.Menu", "SystemColors.MenuBrush"},
                    {"KnownColor.MenuText", "SystemColors.MenuTextBrush"},
                    {"KnownColor.ScrollBar", "SystemColors.ScrollBarBrush"},
                    {"KnownColor.Window", "SystemColors.WindowBrush"},
                    {"KnownColor.WindowFrame", "SystemColors.WindowFrameBrush"},
                    {"KnownColor.WindowText", "SystemColors.WindowTextBrush"},
                    {"KnownColor.ButtonFace", "SystemColors.ControlBrush"}, // Missing in SystemColors.
                    {"KnownColor.ButtonHighlight", "SystemColors.HighlightTextBrush"}, // Missing in SystemColors.
                    {"KnownColor.ButtonShadow", "SystemColors.ControlDarkBrush"}, // Missing in SystemColors.
                    {"KnownColor.GradientActiveCaption", "SystemColors.GradientActiveCaptionBrush"},
                    {"KnownColor.GradientInactiveCaption", "SystemColors.GradientInactiveCaptionBrush"},
                    {"KnownColor.MenuBar", "SystemColors.MenuBarBrush"},
                    {"KnownColor.MenuHighlight", "SystemColors.MenuHighlightBrush"},
                };

            // Lookup of virtual methods that have been renamed.
            // The key is the name of the old method.
            // The value is the signature of the new method.
            private Dictionary<string, string> _renamedVirtualMethods =
                new Dictionary<string, string>
                {
                    {"GetMinMaxValues", "public override void OnCalculateMinMax()"},
                    {"Plot", "protected override void OnRender(ChartControl chartControl, ChartScale chartScale)"},
                    {
                        "OnConnectionStatus",
                        "protected override void OnConnectionStatusUpdate(ConnectionStatus orderStatus, ConnectionStatus priceStatus)"
                    },
                    {
                        "OnOrderUpdate",
                        "protected override void OnOrderUpdate(Order order, double limitPrice, double stopPrice, int quantity, int filled, double averageFillPrice, OrderState orderState, DateTime time, ErrorCode error, string comment)"
                    },
                    {
                        "OnExecution",
                        "protected override void OnExecutionUpdate(Execution execution, string executionId, double price, int quantity, MarketPosition marketPosition, string orderId, DateTime time)"
                    },
                    {
                        "OnPositionUpdate",
                        "protected override void OnPositionUpdate(Position position, double averagePrice, int quantity, MarketPosition marketPosition)"
                    },
                };

            // Lookup of properties that were renamed.
            // The key is the old property name.
            // The value is the new property name.
            //
            // NOTE:
            //   If the renamed property is a boolean that is now has the reversed meaning (e.g.
            //   ExcludeTradeHistoryInBacktest -> IncludeTradeHistoryInBacktest), the appropriate entry
            //   needs to be added in the _replacedBooleanProperties field below - see the comments there.
            private Dictionary<string, string> _renamedProperties =
                new Dictionary<string, string>
                {
                    {"CalculateOnBarClose", "Calculate"},
                    {"Overlay", "IsOverlay"},
                    {"AutoScale", "IsAutoScale"},
                    {"ChartOnly", "IsChartOnly"},
                    {"Unmanaged", "IsUnmanaged"},
                    {"FirstTickOfBar", "IsFirstTickOfBar"},
                    {"LinesConfigurable", "AreLinesConfigurable"},
                    {"PlotsConfigurable", "ArePlotsConfigurable"},
                    {"MaxRestartAttempts", "NumberRestartAttempts"},
                    {"MaxRestartMinutes", "RestartsWithinMinutes"},
                    {"RestartDelaySeconds", "DisconnectDelaySeconds"},
                    {"ExitOnClose", "IsExitOnSessionCloseStrategy"},
                    {"ExitOnCloseSeconds", "ExitOnSessionCloseSeconds"},
                    {"ForceMaximumBarsLookBack256", "ForcePlotsMaximumBarsLookBackInfinite"},
                    {"ExcludeTradeHistoryInBacktest", "IncludeTradeHistoryInBacktest"},
                    {"DataSeriesConfigurable", "IsDataSeriesRequired"},
                    {"BackColor", "BackBrush"},
                    {"BackColorSeries", "BackBrushesAll"},
                    {"BackColorAll", "BackBrushAll"},
                    {"BackColorAllSeries", "BackBrushesAll"},
                    {"BarColor", "BarBrush"},
                    {"BarColorSeries", "BarBrushes"},
                    {"CandleOutlineColor", "CandleOutlineBrush"},
                    {"CandleOutlineColorSeries", "CandleOutlineBrushes"},
                    {"PlotColors", "PlotBrushes"},
                    {"FirstBarIndexPainted", "ChartBars.FromIndex"},
                    {"LastBarIndexPainted", "ChartBars.ToIndex"},
                    {"BarsRequired", "BarsRequiredToPlot"},
                    {"VerticalGridLines", "DrawVerticalGridLines"},
                    {"SessionBreakLines", "IsTradingHoursBreakLineVisible"},
                    {"InHitTest", "IsInHitTest"},
                };

            // Lookup of properties that were removed.  These will be commented out during conversion.
            private List<string> _removedProperties =
                new List<string>
                {
                    {"PriceTypeSupported"},
                };

            // Lookup of methods that were removed.  These will be commented out during conversion.
            private List<string> _removedMethods =
                new List<string>
                {
                    {"ResetAlerts"},
                };

            // Lookup of boolean properties that were replaced.
            // The key is the old property name.
            // The value contains: replacement when checking for true,replacement when checking for false.
            //   e.g. if (CalculateOnBarClose && !Historical) ...; would change to
            //        if ((Calculate == Calculate.OnBarClose) && (State != State.Historical)) ...;
            //
            // There is a hack that treats any replacement that doesn't contain a '(' as a reversed replacement.
            //   e.g. ExcludeTradeHistoryInBacktest was renamed to IncludeTradeHistoryInBacktest.
            //        The meaning is reversed and since the replacement does not contain a '(':
            //            ExcludeTradeHistoryInBacktest = someBooleanValue;
            //        Will be changed to:
            //            IncludeTradeHistoryInBacktest = !(someBooleanValue);
            //
            private Dictionary<string, string> _replacedBooleanProperties =
                new Dictionary<string, string>
                {
                    {"CalculateOnBarClose", "(Calculate == Calculate.OnBarClose),(Calculate != Calculate.OnBarClose)"},
                    {"Calculate", "(Calculate == Calculate.OnBarClose),(Calculate != Calculate.OnBarClose)"},
                    {"Historical", "(State == State.Historical),(State != State.Historical)"},
                    {
                        "ForceMaximumBarsLookBack256",
                        "!ForcePlotsMaximumBarsLookBackInfinite,ForcePlotsMaximumBarsLookBackInfinite"
                    },
                    {"ExcludeTradeHistoryInBacktest", "!IncludeTradeHistoryInBacktest,IncludeTradeHistoryInBacktest"},
                };

            // List of Draw functions to convert.  Convert: DrawArc("tag", ...) to Draw.Arc(this, "tag", ...).
            private List<string> _drawMethods =
                new List<string>
                {
                    "DrawArc",
                    "DrawAndrewsPitchfork",
                    "DrawArrowDown",
                    "DrawArrowUp",
                    "DrawArrowLine",
                    "DrawDiamond",
                    "DrawDot",
                    "DrawEllipse",
                    "DrawExtendedLine",
                    "DrawFibonacciCircle",
                    "DrawFibonacciExtensions",
                    "DrawFibonacciRetracements",
                    "DrawFibonacciTimeExtensions",
                    "DrawGannFan",
                    "DrawHorizontalLine",
                    "DrawLine",
                    "DrawRay",
                    "DrawRectangle",
                    "DrawRegion",
                    "DrawRegressionChannel",
                    "DrawRuler",
                    "DrawSquare",
                    "DrawText",
                    "DrawTextFixed",
                    "DrawTriangle",
                    "DrawTrendChannel",
                    "DrawTriangle",
                    "DrawTriangleDown",
                    "DrawTriangleUp",
                    "DrawVerticalLine",
                };

            public string ConvertClassText(string text)
            {
                var watch = Stopwatch.StartNew();
                // Remove the generated code.
                text = Regex.Replace(text, @"#region NinjaScript generated code.*", "", RegexOptions.Singleline);

                var rewriter = new Nt8Rewriter(
                    _renamedConstructors,
                    _renamedMembers,
                    _renamedVirtualMethods,
                    _removedMethods,
                    _removedProperties,
                    _replacedBooleanProperties,
                    _drawMethods);

                var tree = (CSharpSyntaxTree)CSharpSyntaxTree.ParseText(text);
                SyntaxNode root = tree.GetRoot();

                root = AddUsings(root);
                root = UpdateFontConstructors(root);
                root = UpdateGetProfitLossInvocation(root);
                root = rewriter.Visit(root);
                root = ReplaceTypes(root, _renamedTypes);
                root = ReplacePropertyNames(root, _renamedProperties);
                root = ReplaceSeriesGets(root);
                root = UpdateMinMaxMethod(root);
                root = UpdateOnRenderMethod(root);

                root = AddSpecialMembers(
                    root,
                    rewriter.ClassName,
                    rewriter.ClassDescription,
                    rewriter.HasInitializeMethod,
                    rewriter.HasOnStartupMethod,
                    rewriter.HasOnTerminationMethod,
                    rewriter.HasBrushFromNameCall,
                    rewriter.HasBrushFromArgbCall);
                return root.ToFullString();
            }
            private SyntaxNode AddUsings(SyntaxNode node)
            {
                var result = (CompilationUnitSyntax)node;

                List<UsingDirectiveSyntax> usings = CreateSyntaxNode(_usingsText)
                    .DescendantNodes()
                    .OfType<UsingDirectiveSyntax>()
                    .ToList();

                SyntaxTriviaList oldLeading = SyntaxFactory.TriviaList();

                foreach (UsingDirectiveSyntax usingDirective in result.Usings)
                {
                    bool hasNonwhitespaceTrivia = usingDirective
                        .GetLeadingTrivia()
                        .Any(t => !t.IsKind(SyntaxKind.WhitespaceTrivia) && !t.IsKind(SyntaxKind.EndOfLineTrivia));

                    if (hasNonwhitespaceTrivia)
                        oldLeading = oldLeading.AddRange(usingDirective.GetLeadingTrivia());
                }

                usings[0] = usings[0].WithLeadingTrivia(oldLeading);

                return result.WithUsings(SyntaxFactory.List<UsingDirectiveSyntax>(usings));
            }

            private SyntaxNode ReplaceTypes(SyntaxNode node, Dictionary<string, string> typeNameLookup)
            {
                IEnumerable<NameSyntax> types = GetDescendantNodesOfTypeWithPropertyName<NameSyntax>(node, "Type",
                    "ReturnType");

                IEnumerable<NameSyntax> typeArguments = node
                    .DescendantNodes()
                    .OfType<TypeArgumentListSyntax>()
                    .SelectMany(a => a.DescendantNodes().OfType<NameSyntax>());

                IEnumerable<NameSyntax> typesToReplace = types
                    .Union(typeArguments)
                    .Where(n => typeNameLookup.ContainsKey(n.ToString()));

                var replacements = new Dictionary<SyntaxNode, SyntaxNode>();

                foreach (NameSyntax identifierName in typesToReplace)
                {
                    string newName = typeNameLookup[identifierName.ToString()];
                    SyntaxNode newIdentifierName = SyntaxFactory.ParseName(newName).WithTriviaFrom(identifierName);
                    replacements.Add(identifierName, newIdentifierName);
                }

                return node.ReplaceNodes(replacements.Keys, (n, _) => replacements[n]);
            }

            private SyntaxNode ReplacePropertyNames(SyntaxNode node, Dictionary<string, string> propertyNameLookup)
            {
                // HACK: Find properties by looking for identifiers that are not Types. This is simplier than
                //       adding all the different cases for different constructs.
                //
                // Should work for most cases but using semantic analysis will be better.
                IEnumerable<NameSyntax> identifiers = node
                    .DescendantNodes()
                    .OfType<NameSyntax>()
                    .Where(n => propertyNameLookup.ContainsKey(n.ToString()));

                IEnumerable<NameSyntax> types = GetDescendantNodesOfTypeWithPropertyName<NameSyntax>(node, "Type",
                    "ReturnType");
                IEnumerable<NameSyntax> propertyIdentifiers = identifiers.Except(types);
                var replacements = new Dictionary<SyntaxNode, SyntaxNode>();

                foreach (NameSyntax identifierName in propertyIdentifiers)
                {
                    string newName = propertyNameLookup[identifierName.ToString()];
                    SyntaxNode newIdentifierName = SyntaxFactory.ParseName(newName).WithTriviaFrom(identifierName);
                    replacements.Add(identifierName, newIdentifierName);
                }

                return node.ReplaceNodes(replacements.Keys, (n, _) => replacements[n]);
            }

            private SyntaxNode ReplaceSeriesGets(SyntaxNode node)
            {
                while (true)
                {
                    InvocationExpressionSyntax invocation = node
                        .DescendantNodes()
                        .OfType<InvocationExpressionSyntax>()
                        .Where(d => ((d.Expression as MemberAccessExpressionSyntax)?.Name.ToString() == "Get"))
                        .FirstOrDefault();

                    if (invocation == null) return node;

                    var expression = (MemberAccessExpressionSyntax)invocation.Expression;
                    var memberName = expression.Expression as NameSyntax;
                    ArgumentSyntax indexArgument = invocation.ArgumentList.Arguments[0];

                    if (memberName == null) return node;

                    BracketedArgumentListSyntax brackets = SyntaxFactory
                        .BracketedArgumentList(SyntaxFactory.SingletonSeparatedList<ArgumentSyntax>(indexArgument))
                        .WithOpenBracketToken(SyntaxFactory.Token(SyntaxKind.OpenBracketToken))
                        .WithCloseBracketToken(SyntaxFactory.Token(SyntaxKind.CloseBracketToken));

                    ElementAccessExpressionSyntax elementAccess = SyntaxFactory
                        .ElementAccessExpression(memberName, brackets);

                    var newNode = invocation.Parent
                        .ReplaceNode(invocation, elementAccess)
                        .WithTrailingTrivia(invocation.GetTrailingTrivia());

                    node = node.ReplaceNode(invocation.Parent, newNode);
                }
            }

            private SyntaxNode UpdateMinMaxMethod(SyntaxNode node)
            {
                MethodDeclarationSyntax minMaxMethod = node
                    .DescendantNodes()
                    .OfType<MethodDeclarationSyntax>()
                    .FirstOrDefault(n => n.Identifier.ToString() == "OnCalculateMinMax");

                if (minMaxMethod != null)
                {
                    node = ReplacePropertyNames(
                        node,
                        new Dictionary<string, string>
                        {
                            {"min", "MinValue"},
                            {"max", "MaxValue"},
                        });
                }

                return node;
            }

            private SyntaxNode UpdateOnRenderMethod(SyntaxNode node)
            {
                // Re-parse since some identifiers have been hacked with comments.
                node = SyntaxFactory.ParseSyntaxTree(node.ToFullString()).GetRoot();

                MethodDeclarationSyntax onRenderMethod = node
                    .DescendantNodes()
                    .OfType<MethodDeclarationSyntax>()
                    .FirstOrDefault(n => n.Identifier.ToString() == "OnRender");

                if (onRenderMethod == null) return node;

                var newOnRenderMethod = (MethodDeclarationSyntax)ReplacePropertyNames(
                    onRenderMethod,
                    new Dictionary<string, string>
                    {
                        {"ChartControl", "chartControl"},
                    });

                var rewriter = new Nt8Rewriter(
                    null,
                    new Dictionary<string, string>
                    {
                        {"Bars.BarsData.ChartStyle.BarWidth", "chartControl.BarWidth"},
                        {"chartControl.GetXByBarIdx", "chartControl.GetXByBarIndex"},
                        {"chartControl.GetYByValue", "chartScale.GetYByValue"},
                    },
                    null,
                    null,
                    null,
                    null,
                    null);

                node = node.ReplaceNode(onRenderMethod, rewriter.Visit(newOnRenderMethod));
                return node;
            }

            private SyntaxNode UpdateGetProfitLossInvocation(SyntaxNode node)
            {
                InvocationExpressionSyntax getProfitLoss = node
                    .DescendantNodes()
                    .OfType<InvocationExpressionSyntax>()
                    .FirstOrDefault(n => n.Expression.ToString() == "Position.GetProfitLoss");

                if (getProfitLoss != null)
                {
                    SeparatedSyntaxList<ArgumentSyntax> oldArguments = getProfitLoss.ArgumentList.Arguments;
                    SeparatedSyntaxList<ArgumentSyntax> newArguments = oldArguments;
                    newArguments = newArguments.RemoveAt(1);
                    newArguments = newArguments.Insert(0, oldArguments[1]);
                    newArguments = newArguments.Replace(newArguments[0], newArguments[0].WithTriviaFrom(oldArguments[0]));
                    newArguments = newArguments.Replace(newArguments[1], newArguments[1].WithTriviaFrom(oldArguments[1]));
                    SyntaxToken newSeparator = newArguments.GetSeparator(0).WithTriviaFrom(oldArguments.GetSeparator(0));
                    newArguments = newArguments.ReplaceSeparator(newArguments.GetSeparator(0), newSeparator);

                    ArgumentListSyntax newArgumentList = getProfitLoss.ArgumentList.WithArguments(newArguments);
                    node = node.ReplaceNode(getProfitLoss, getProfitLoss.WithArgumentList(newArgumentList));
                }

                return node;
            }

            private SyntaxNode UpdateFontConstructors(SyntaxNode node)
            {
                IEnumerable<ObjectCreationExpressionSyntax> simpleFonts = node
                    .DescendantNodes()
                    .OfType<ObjectCreationExpressionSyntax>()
                    .Where(f => f.Type.ToString() == "Font")
                    .Where(f => f.ArgumentList.Arguments.Count > 2);

                var replacements = new Dictionary<SyntaxNode, SyntaxNode>();
                var initializations = new List<string>();

                foreach (ObjectCreationExpressionSyntax font in simpleFonts)
                {
                    ArgumentListSyntax newArgumentsList = font
                        .ArgumentList
                        .RemoveNodes(font.ArgumentList.Arguments.Skip(2), SyntaxRemoveOptions.KeepNoTrivia);

                    ObjectCreationExpressionSyntax newFont = font.WithArgumentList(newArgumentsList);
                    ArgumentSyntax fontStyleArgument = font.ArgumentList.Arguments[2];
                    initializations.Clear();

                    bool hasBold = fontStyleArgument
                        .DescendantNodes()
                        .OfType<MemberAccessExpressionSyntax>()
                        .Any(m => m.ToString() == "FontStyle.Bold");

                    bool hasItalic = fontStyleArgument
                        .DescendantNodes()
                        .OfType<MemberAccessExpressionSyntax>()
                        .Any(m => m.ToString() == "FontStyle.Italic");

                    bool hasOtherFontStyle = fontStyleArgument
                        .DescendantNodes()
                        .OfType<MemberAccessExpressionSyntax>()
                        .Any(m => m.ToString() != "FontStyle.Bold" && m.ToString() != "FontStyle.Italic");

                    if (hasBold)
                        initializations.Add("Bold = true");

                    if (hasItalic)
                        initializations.Add("Italic = true");

                    if (initializations.Count > 0)
                    {
                        newFont = (ObjectCreationExpressionSyntax)SyntaxFactory
                            .ParseExpression(newFont + $" {{ {string.Join(", ", initializations)} }}");
                    }

                    if (hasOtherFontStyle || font.ArgumentList.Arguments.Count > 3)
                    {
                        SyntaxTrivia comment = SyntaxFactory.Comment($"/*{font.ToString()}*/");
                        newFont = newFont.WithTrailingTrivia(newFont.GetTrailingTrivia().Add(comment));
                    }

                    replacements.Add(font, newFont);
                }

                return node.ReplaceNodes(replacements.Keys, (n, _) => replacements[n]);
            }

            private SyntaxNode CreateOnStateChangeMethod(
                string className,
                string classDescription,
                bool hasInitializeMethod,
                bool hasOnStartupMethod,
                bool hasOnTerminationMethod
                )
            {
                string methodText = @"
        protected override void OnStateChange()
        {
            switch (State)
            {
                case State.SetDefaults:
                    Name = $name$;
                    Description = $description$;
                    $initialize$
                    break;$startupcase$$terminationcase$
             }
        }
";

                string startupCaseText = @"

                case State.DataLoaded:
                    OnStartUp();
                    break;";

                string terminationCaseText = @"

                case State.Terminated:
                    OnTermination();
                    break;";

                methodText = methodText.Replace("$name$", className);
                methodText = methodText.Replace("$description$", classDescription);
                methodText = methodText.Replace("$initialize$", hasInitializeMethod ? "Initialize();" : "");
                methodText = methodText.Replace("$startupcase$", hasOnStartupMethod ? startupCaseText : "");
                methodText = methodText.Replace("$terminationcase$", hasOnTerminationMethod ? terminationCaseText : "");

                return CreateSyntaxNode(methodText)
                    .DescendantNodes()
                    .OfType<MethodDeclarationSyntax>()
                    .First();
            }

            private IEnumerable<SyntaxNode> CreateBrushFromNameMembers()
            {
                string methodsText = @"public class Temp {

        private static SolidColorBrush BrushFromName(string name)
        {
            try
            {
                return new SolidColorBrush((Color)ColorConverter.ConvertFromString(name));
            }
            catch
            {
            }
        
            Type brushType = typeof(SystemColors);
            PropertyInfo property = brushType.GetProperty(name, BindingFlags.Public | BindingFlags.Static);
            return property != null ? (SolidColorBrush)property.GetValue(null) : Brushes.Transparent;
        }

}";

                return SyntaxFactory
                    .ParseSyntaxTree(methodsText)
                    .GetRoot()
                    .DescendantNodes()
                    .OfType<MethodDeclarationSyntax>();
            }

            private IEnumerable<SyntaxNode> CreateBrushFromArgbMembers()
            {
                string methodsText = @"public class Temp {

        private static SolidColorBrush BrushFromArgb(int argb)
        {
            return new SolidColorBrush(Color.FromArgb(
                (byte)(argb >> 24),
                (byte)(argb >> 16),
                (byte)(argb >> 8),
                (byte)(argb)));
        }

        private static SolidColorBrush BrushFromArgb(int alpha, Brush baseBrush)
        {
            var brush = (SolidColorBrush)baseBrush;

            return new SolidColorBrush(Color.FromArgb(
                (byte)alpha,
                brush.Color.R,
                brush.Color.G,
                brush.Color.B));
        }
}";

                return SyntaxFactory
                    .ParseSyntaxTree(methodsText)
                    .GetRoot()
                    .DescendantNodes()
                    .OfType<MethodDeclarationSyntax>();
            }

            private SyntaxNode AddSpecialMembers(
                SyntaxNode node,
                string className,
                string classDescription,
                bool hasInitializeMethod,
                bool hasOnStartupMethod,
                bool hasOnTerminationMethod,
                bool hasBrushFromNameCall,
                bool hasBrushFromArgbCall
                )
            {
                SyntaxNode referenceNode = null;
                ClassDeclarationSyntax classDeclaration = GetFirstDescendantNode<ClassDeclarationSyntax>(node);

                if (classDeclaration != null)
                {
                    SyntaxNode firstMethod = GetFirstChildNode<MethodDeclarationSyntax>(classDeclaration);
                    SyntaxNode lastField = GetLastChildNode<FieldDeclarationSyntax>(classDeclaration);
                    SyntaxNode lastProperty = GetLastChildNode<PropertyDeclarationSyntax>(classDeclaration);

                    referenceNode = firstMethod ?? lastField ?? lastProperty;
                }

                if (referenceNode == null) return node;

                SyntaxNode onStateChangeMethod = CreateOnStateChangeMethod(
                    className,
                    classDescription,
                    hasInitializeMethod,
                    hasOnStartupMethod,
                    hasOnTerminationMethod);

                var newNodes = new List<SyntaxNode>();
                newNodes.Add(onStateChangeMethod);

                if (hasBrushFromNameCall)
                    newNodes.AddRange(CreateBrushFromNameMembers());

                if (hasBrushFromArgbCall)
                    newNodes.AddRange(CreateBrushFromArgbMembers());

                return node.InsertNodesAfter(referenceNode, newNodes);
            }
        }

        public class Nt8Rewriter : CSharpSyntaxRewriter
        {
            public Nt8Rewriter(
                Dictionary<string, string> renamedConstructorLookup,
                Dictionary<string, string> renamedMembersLookup,
                Dictionary<string, string> renamedVirtualMethodLookup,
                List<string> removedMethods,
                List<string> removedProperties,
                Dictionary<string, string> replacedBooleanPropertiesLookup,
                List<string> drawMethods)
                : base(visitIntoStructuredTrivia: true)
            {
                _constructorLookup = renamedConstructorLookup ?? new Dictionary<string, string>();
                _membersLookup = renamedMembersLookup ?? new Dictionary<string, string>();
                _virtualMethodLookup = renamedVirtualMethodLookup ?? new Dictionary<string, string>();
                _removedMethodsLookup = new HashSet<string>(removedMethods ?? new List<string>());
                _removedPropertiesLookup = new HashSet<string>(removedProperties ?? new List<string>());
                _replacedBooleanPropertiesLookup = replacedBooleanPropertiesLookup ?? new Dictionary<string, string>();
                _drawMethodsLookup = new HashSet<string>(drawMethods ?? new List<string>());
            }

            public bool IsIndicator { get; private set; }
            public string ClassName { get; private set; }
            public string ClassDescription { get; private set; }
            public bool HasInitializeMethod { get; private set; }
            public bool HasOnStartupMethod { get; private set; }
            public bool HasOnTerminationMethod { get; private set; }
            public bool HasBrushFromNameCall { get; private set; }
            public bool HasBrushFromArgbCall { get; private set; }

            public override SyntaxNode VisitNamespaceDeclaration(NamespaceDeclarationSyntax node)
            {
                IsIndicator = node.Name.ToString() == "NinjaTrader.Indicator";

                node = (NamespaceDeclarationSyntax)base.VisitNamespaceDeclaration(node);

                string newNamespaceNameText = IsIndicator
                    ? "NinjaTrader.NinjaScript.Indicators"
                    : "NinjaTrader.NinjaScript.Strategies";

                NameSyntax newNamespaceName = SyntaxFactory
                    .ParseName(newNamespaceNameText)
                    .WithTrailingTrivia(node.Name.GetTrailingTrivia());

                return node.WithName(newNamespaceName);
            }

            public override SyntaxNode VisitClassDeclaration(ClassDeclarationSyntax node)
            {
                node = (ClassDeclarationSyntax)base.VisitClassDeclaration(node);

                var baseType = node.BaseList?.ChildNodes().OfType<SimpleBaseTypeSyntax>().FirstOrDefault();

                if (baseType != null)
                {
                    var nameIdentifier = baseType.Type as NameSyntax;
                    string baseClassName = nameIdentifier.ToString();
                    string newBaseClassNameText = IsIndicator ? "Indicator" : "Strategy";

                    if (baseClassName == "Indicator.Indicator"
                        || baseClassName == "NinjaTrader.Indicator.Indicator"
                        || baseClassName == "Strategy.Strategy"
                        || baseClassName == "NinjaTrader.Strategy.Strategy")
                    {
                        var newBaseType = baseType
                            .WithType(SyntaxFactory.ParseName(newBaseClassNameText))
                            .WithTriviaFrom(baseType);

                        var newBaseList = node.BaseList.ReplaceNode(baseType, newBaseType);
                        node = node.WithBaseList(newBaseList);
                    }
                }

                AttributeSyntax displayNameAttribute = GetAttribute(node, "DisplayName")
                                                       ?? GetAttribute(node, "Design.DisplayName")
                                                       ?? GetAttribute(node, "Gui.Design.DisplayName");

                AttributeSyntax descriptionAttribute = GetAttribute(node, "Description");
                ClassName = GetAttributeArgument(displayNameAttribute, 0) ?? $"\"{node.Identifier.ValueText}\"";
                ClassDescription = GetAttributeArgument(descriptionAttribute, 0) ?? $"\"{node.Identifier.ValueText}\"";

                if (displayNameAttribute != null || descriptionAttribute != null)
                    node = ExcludeAttributes(node, displayNameAttribute, descriptionAttribute);

                return node;
            }

            public override SyntaxNode VisitMethodDeclaration(MethodDeclarationSyntax node)
            {
                node = (MethodDeclarationSyntax)base.VisitMethodDeclaration(node);

                string methodName = node.Identifier.ValueText;
                string newMethodSignature;
                bool isSpecialMethod = false;

                if (methodName == "Initialize")
                {
                    isSpecialMethod = true;
                    HasInitializeMethod = true;
                }
                else if (methodName == "OnStartUp")
                {
                    isSpecialMethod = true;
                    HasOnStartupMethod = true;
                }
                else if (methodName == "OnTermination")
                {
                    isSpecialMethod = true;
                    HasOnTerminationMethod = true;
                }

                if (isSpecialMethod)
                {
                    var protectedKeyword = node.DescendantTokens().First(t => t.IsKind(SyntaxKind.ProtectedKeyword));

                    // Replace the protected and override modifiers with private.
                    var privateKeyword = SyntaxFactory
                        .Token(SyntaxKind.PrivateKeyword)
                        .WithTriviaFrom(protectedKeyword);

                    node = node.WithModifiers(SyntaxFactory.TokenList(new[] { privateKeyword }));
                }
                else if (_virtualMethodLookup.TryGetValue(methodName, out newMethodSignature))
                {
                    var newMethod = (MethodDeclarationSyntax)SyntaxFactory
                        .ParseSyntaxTree(newMethodSignature + "{}")
                        .GetRoot()
                        .DescendantNodes()
                        .OfType<MethodDeclarationSyntax>()
                        .First()
                        .WithBody(node.Body)
                        .WithTriviaFrom(node);

                    ParameterListSyntax newParameterList = newMethod.ParameterList.WithTriviaFrom(node);
                    node = newMethod.WithParameterList(newParameterList);
                }

                return node;
            }

            public override SyntaxNode VisitBinaryExpression(BinaryExpressionSyntax node)
            {
                node = (BinaryExpressionSyntax)base.VisitBinaryExpression(node);

                node = node.WithLeft(ApplyReplacedBooleanProperty(node.Left));
                node = node.WithRight(ApplyReplacedBooleanProperty(node.Right));
                return node;
            }

            public override SyntaxNode VisitIfStatement(IfStatementSyntax node)
            {
                node = (IfStatementSyntax)base.VisitIfStatement(node);

                var identifierName = node.Condition as NameSyntax;
                var urinaryExpression = node.Condition as PrefixUnaryExpressionSyntax;
                var memberAccessExpression = node.Condition as MemberAccessExpressionSyntax;

                if (identifierName != null || urinaryExpression != null || memberAccessExpression != null)
                    node = node.WithCondition(ApplyReplacedBooleanProperty(node.Condition));

                return node;
            }

            public override SyntaxNode VisitConditionalExpression(ConditionalExpressionSyntax node)
            {
                node = (ConditionalExpressionSyntax)base.VisitConditionalExpression(node);

                var identifierName = node.Condition as NameSyntax;
                var urinaryExpression = node.Condition as PrefixUnaryExpressionSyntax;
                var memberAccessExpression = node.Condition as MemberAccessExpressionSyntax;

                if (identifierName != null || urinaryExpression != null || memberAccessExpression != null)
                    node = node.WithCondition(ApplyReplacedBooleanProperty(node.Condition));

                return node;
            }

            public override SyntaxNode VisitParenthesizedExpression(ParenthesizedExpressionSyntax node)
            {
                node = (ParenthesizedExpressionSyntax)base.VisitParenthesizedExpression(node);

                var identifierName = node.Expression as NameSyntax;
                var urinaryExpression = node.Expression as PrefixUnaryExpressionSyntax;
                var memberAccessExpression = node.Expression as MemberAccessExpressionSyntax;

                if (identifierName != null || urinaryExpression != null || memberAccessExpression != null)
                    node = node.WithExpression(ApplyReplacedBooleanProperty(node.Expression));

                return node;
            }

            public override SyntaxNode VisitEqualsValueClause(EqualsValueClauseSyntax node)
            {
                node = (EqualsValueClauseSyntax)base.VisitEqualsValueClause(node);

                node = node.WithValue(ApplyReplacedBooleanProperty(node.Value));
                return node;
            }

            public override SyntaxNode VisitObjectCreationExpression(ObjectCreationExpressionSyntax node)
            {
                node = (ObjectCreationExpressionSyntax)base.VisitObjectCreationExpression(node);

                var objectType = node.Type as NameSyntax;

                if (objectType == null) return node;

                string objectTypeName = objectType.ToString();
                string newTypeName;

                if (_constructorLookup.TryGetValue(objectType.ToString(), out newTypeName))
                {
                    TypeSyntax newObjectType = SyntaxFactory
                        .ParseTypeName(newTypeName)
                        .WithTrailingTrivia(objectType.GetTrailingTrivia());

                    node = node.WithType(newObjectType);
                }
                else if (objectTypeName == "SolidBrush")
                {
                    ArgumentSyntax argument = node.ArgumentList.Arguments[0];
                    var memberAccess = argument.Expression as MemberAccessExpressionSyntax;
                    var invocation = argument.Expression as InvocationExpressionSyntax;
                    // We've hacked the tree so we have to check for this (tree is invalid with this here).
                    var identifier = argument.Expression as IdentifierNameSyntax;
                    int invocationArgumentCount = invocation != null ? invocation.ArgumentList.Arguments.Count : 0;


                    if (memberAccess != null
                        || identifier != null
                        || (invocationArgumentCount != 3 && invocationArgumentCount != 4))
                    {
                        return argument.Expression.WithTriviaFrom(node);
                    }
                }

                return node;
            }

            public override SyntaxNode VisitMemberAccessExpression(MemberAccessExpressionSyntax node)
            {
                node = (MemberAccessExpressionSyntax)base.VisitMemberAccessExpression(node);

                if (!node.IsKind(SyntaxKind.SimpleMemberAccessExpression)) return node;

                string memberName = node.ToString();
                string newName = null;

                if (_membersLookup.TryGetValue(memberName, out newName))
                {
                    if (!newName.Contains('('))
                        return SyntaxFactory.IdentifierName(newName).WithTriviaFrom(node);

                    return SyntaxFactory.ParseExpression(newName).WithTriviaFrom(node);
                }

                var classIdentifier = node.Expression as NameSyntax;

                if (classIdentifier == null || !_membersLookup.TryGetValue(classIdentifier.ToString(), out newName))
                    return node;

                NameSyntax newClassIdentifier = SyntaxFactory
                    .IdentifierName(newName)
                    .WithTriviaFrom(classIdentifier);

                return node.WithExpression(newClassIdentifier);
            }

            private ExpressionStatementSyntax ProcessSeriesSet(ExpressionStatementSyntax node,
                InvocationExpressionSyntax seriesSetInvocation)
            {
                var expression = (MemberAccessExpressionSyntax)seriesSetInvocation.Expression;

                ExpressionSyntax memberName = expression.Expression;
                SyntaxTrivia space = SyntaxFactory.SyntaxTrivia(SyntaxKind.WhitespaceTrivia, " ");
                SyntaxToken equals =
                    SyntaxFactory.Token(SyntaxKind.EqualsToken).WithLeadingTrivia(space).WithTrailingTrivia(space);
                SyntaxToken semicolon = SyntaxFactory.Token(SyntaxKind.SemicolonToken);
                ArgumentSyntax indexArgument = null;
                ArgumentSyntax rvalue = null;

                if (seriesSetInvocation.ArgumentList.Arguments.Count == 1)
                {
                    indexArgument = SyntaxFactory
                        .Argument(SyntaxFactory.LiteralExpression(SyntaxKind.NumericLiteralExpression,
                            SyntaxFactory.Literal("0", 0)));

                    rvalue = seriesSetInvocation.ArgumentList.Arguments[0];
                }
                else if (seriesSetInvocation.ArgumentList.Arguments.Count == 2)
                {
                    indexArgument = seriesSetInvocation.ArgumentList.Arguments[0];
                    rvalue = seriesSetInvocation.ArgumentList.Arguments[1];
                }

                BracketedArgumentListSyntax brackets = SyntaxFactory
                    .BracketedArgumentList(SyntaxFactory.SingletonSeparatedList<ArgumentSyntax>(indexArgument))
                    .WithOpenBracketToken(SyntaxFactory.Token(SyntaxKind.OpenBracketToken))
                    .WithCloseBracketToken(SyntaxFactory.Token(SyntaxKind.CloseBracketToken));

                ElementAccessExpressionSyntax elementAccess = SyntaxFactory
                    .ElementAccessExpression(memberName, brackets);

                AssignmentExpressionSyntax assignment = SyntaxFactory
                    .AssignmentExpression(SyntaxKind.SimpleAssignmentExpression, elementAccess, equals,
                        rvalue.Expression);

                return SyntaxFactory
                    .ExpressionStatement(assignment)
                    .WithSemicolonToken(semicolon)
                    .WithTrailingTrivia(node.GetTrailingTrivia());
            }

            private ExpressionStatementSyntax ProcessAddPlot(ExpressionStatementSyntax node,
                InvocationExpressionSyntax invocation, ObjectCreationExpressionSyntax plot)
            {
                ArgumentListSyntax newArgumentList = plot.ArgumentList;
                int plotArgumentCount = plot.ArgumentList.Arguments.Count;

                if (plotArgumentCount == 2)
                {
                    var plotArgument1 = plot.ArgumentList.Arguments[0].Expression as ObjectCreationExpressionSyntax;

                    if ((plotArgument1?.Type as NameSyntax)?.ToString() == "Stroke")
                    {
                        // Add(new Plot(new Stroke(Brush), "name")) -> AddPlot(new Stroke(Brush), PlotStyle, "name").
                        ArgumentSyntax plotStyleArgument = SyntaxFactory
                            .ParseArgumentList($"(PlotStyle.Line)")
                            .Arguments[0]
                            .WithTriviaFrom(plotArgument1);

                        newArgumentList = plot.ArgumentList
                            .InsertNodesAfter(plot.ArgumentList.Arguments[0], new[] { plotStyleArgument });

                        SyntaxToken firstComma = newArgumentList.Arguments.GetSeparator(0);
                        SyntaxToken secondComma = newArgumentList.Arguments.GetSeparator(1);
                        newArgumentList = newArgumentList.ReplaceToken(firstComma,
                            firstComma.WithTriviaFrom(secondComma));
                    }
                }
                else if (plotArgumentCount == 3)
                {
                    var plotArgument1 = plot.ArgumentList.Arguments[0].Expression as MemberAccessExpressionSyntax;

                    if ((plotArgument1?.Expression as NameSyntax)?.ToString() == "Brushes")
                    {
                        // new Plot(Brush, PlotStyle, "name") -> new Plot(new Stroke(Brush), PlotStyle, "name").
                        ExpressionSyntax strokeArgument = SyntaxFactory
                            .ParseExpression($"new Stroke({plotArgument1.ToString()})")
                            .WithTriviaFrom(plotArgument1);

                        newArgumentList = plot.ArgumentList.ReplaceNode(plotArgument1, strokeArgument);
                    }
                }

                ExpressionSyntax newInvocation = SyntaxFactory
                    .ParseExpression($"AddPlot{newArgumentList.ToString()}")
                    .WithTriviaFrom(invocation);

                return node.WithExpression(newInvocation);
            }

            private ExpressionStatementSyntax ProcessAddLine(ExpressionStatementSyntax node,
                InvocationExpressionSyntax invocation, ObjectCreationExpressionSyntax line)
            {
                if (line.ArgumentList.Arguments.Count == 3)
                {
                    ExpressionSyntax newInvocation = SyntaxFactory
                        .ParseExpression($"AddLine{line.ArgumentList.ToString()}")
                        .WithTriviaFrom(invocation);

                    node = node.ReplaceNode(invocation, newInvocation);
                }

                return node;
            }

            private ExpressionStatementSyntax ProcessAddIndicator(ExpressionStatementSyntax node,
                InvocationExpressionSyntax invocation, ArgumentSyntax argument)
            {
                var newInvocation = invocation
                    .WithExpression(SyntaxFactory.ParseExpression("AddChartIndicator"))
                    .WithTriviaFrom(invocation.Expression);

                return node.WithExpression(newInvocation);
            }

            private ExpressionStatementSyntax ProcessAddDataSeries(ExpressionStatementSyntax node,
                InvocationExpressionSyntax invocation)
            {
                var newInvocation = invocation
                    .WithExpression(SyntaxFactory.ParseExpression("AddDataSeries"))
                    .WithTriviaFrom(invocation.Expression);

                return node.WithExpression(newInvocation);
            }

            public override SyntaxNode VisitExpressionStatement(ExpressionStatementSyntax node)
            {
                node = (ExpressionStatementSyntax)base.VisitExpressionStatement(node);

                InvocationExpressionSyntax seriesSetInvocation = node
                    .DescendantNodes()
                    .OfType<InvocationExpressionSyntax>()
                    .Where(d => ((d.Expression as MemberAccessExpressionSyntax)?.Name.ToString() == "Set"))
                    .FirstOrDefault();

                if (seriesSetInvocation != null)
                    return ProcessSeriesSet(node, seriesSetInvocation);

                var invocation = node.Expression as InvocationExpressionSyntax;

                if (invocation == null) return node;

                if ((invocation.Expression as NameSyntax)?.ToString() == "Add")
                {
                    if (invocation.ArgumentList.Arguments.Count == 1)
                    {
                        var argument = invocation.ArgumentList.Arguments[0].Expression as ObjectCreationExpressionSyntax;

                        switch ((argument?.Type as NameSyntax)?.ToString())
                        {
                            case "Plot":
                                return ProcessAddPlot(node, invocation, argument);

                            case "Line":
                                return ProcessAddLine(node, invocation, argument);

                            default:
                                return ProcessAddIndicator(node, invocation, invocation.ArgumentList.Arguments[0]);
                        }
                    }
                    else
                    {
                        return ProcessAddDataSeries(node, invocation);
                    }
                }

                return node;
            }

            private SyntaxNode ProcessSolidColorBrush(InvocationExpressionSyntax node, string name)
            {
                if (name == "Color.FromName" || name == "Brushes.FromName")
                {
                    HasBrushFromNameCall = true;

                    var newInvocation = (InvocationExpressionSyntax)SyntaxFactory
                        .ParseExpression("BrushFromName()")
                        .WithTriviaFrom(node);

                    return newInvocation.WithArgumentList(node.ArgumentList);
                }

                if (name != "Color.FromArgb" && name != "Brushes.FromArgb") return node;

                int argumentCount = node.ArgumentList.Arguments.Count;

                if (argumentCount == 1 || argumentCount == 2)
                {
                    HasBrushFromArgbCall = true;

                    var newInvocation = (InvocationExpressionSyntax)SyntaxFactory
                        .ParseExpression("BrushFromArgb()")
                        .WithTriviaFrom(node);

                    return newInvocation.WithArgumentList(node.ArgumentList);
                }

                string newName = argumentCount == 3
                    ? "Color.FromRgb"
                    : argumentCount == 4 ? "Color.FromArgb" : null;

                if (newName != null)
                {
                    NameSyntax newExpression = SyntaxFactory
                        .ParseName(newName)
                        .WithTriviaFrom(node.Expression);

                    node = node.WithExpression(newExpression);
                }

                return node;
            }

            public override SyntaxNode VisitInvocationExpression(InvocationExpressionSyntax node)
            {
                node = (InvocationExpressionSyntax)base.VisitInvocationExpression(node);

                var methodIdentifier = node.Expression as NameSyntax;
                string methodName = methodIdentifier?.ToString();
                var memberAccess = node.Expression as MemberAccessExpressionSyntax;
                string newName;

                if (methodIdentifier != null && _drawMethodsLookup.Contains(methodName))
                {
                    var newMethodIdentifier = SyntaxFactory
                        .ParseName("Draw." + methodName.Substring(4))
                        .WithTriviaFrom(methodIdentifier);

                    var thisArgument = SyntaxFactory
                        .Argument(SyntaxFactory.ThisExpression())
                        .WithTriviaFrom(node.ArgumentList.Arguments[0]);

                    var newInvocationArgumentsList = node
                        .ArgumentList
                        .WithArguments(node.ArgumentList.Arguments.Insert(0, thisArgument));

                    node = node
                        .WithExpression(newMethodIdentifier)
                        .WithArgumentList(newInvocationArgumentsList);
                }
                else if (methodIdentifier != null && _membersLookup.TryGetValue(methodName, out newName))
                {
                    if (!newName.Contains('('))
                        return node.WithExpression(SyntaxFactory.IdentifierName(newName).WithTriviaFrom(node.Expression));

                    return SyntaxFactory.ParseExpression(newName).WithTriviaFrom(node);
                }
                else if (_removedMethodsLookup.Contains(methodName))
                {
                    var comment = node
                        .GetLeadingTrivia()
                        .Add(SyntaxFactory.Comment($"/* NT8 REMOVED: {node.ToString()} */"));

                    return SyntaxFactory
                        .ParseExpression("")
                        .WithLeadingTrivia(SyntaxFactory.TriviaList(comment))
                        .WithTrailingTrivia(node.GetTrailingTrivia());
                }
                else if (memberAccess != null)
                {
                    if (_membersLookup.TryGetValue(memberAccess.Name.ToString(), out newName))
                    {
                        MemberAccessExpressionSyntax newMemberAccess = memberAccess
                            .WithName(SyntaxFactory.IdentifierName(newName).WithTriviaFrom(memberAccess.Name));

                        return node.WithExpression(newMemberAccess);
                    }

                    string name = memberAccess.ToString();

                    if (name == "Color.FromKnownColor" || name == "Brushes.FromKnownColor")
                        return node.ArgumentList.Arguments[0].Expression;

                    return ProcessSolidColorBrush(node, name);
                }

                return node;
            }

            public override SyntaxNode VisitAssignmentExpression(AssignmentExpressionSyntax node)
            {
                node = (AssignmentExpressionSyntax)base.VisitAssignmentExpression(node);

                var variable = node.Left as NameSyntax;

                if (variable == null) return node;

                string variableName = variable.ToString();
                bool isSimpleAssignment = node.IsKind(SyntaxKind.SimpleAssignmentExpression);
                string replacement;

                if (variableName == "CalculateOnBarClose" && isSimpleAssignment)
                {
                    if (node.Right.IsKind(SyntaxKind.TrueLiteralExpression))
                        node = node.WithRight(SyntaxFactory.ParseExpression("Calculate.OnBarClose"));
                    else if (node.Right.IsKind(SyntaxKind.FalseLiteralExpression))
                        node = node.WithRight(SyntaxFactory.ParseExpression("Calculate.OnEachTick"));
                }
                else if (_removedPropertiesLookup.Contains(variableName))
                {
                    var comment = node
                        .GetLeadingTrivia()
                        .Add(SyntaxFactory.Comment($"/* NT8 REMOVED: {node.ToString()} */"));

                    return SyntaxFactory
                        .ParseExpression("")
                        .WithLeadingTrivia(SyntaxFactory.TriviaList(comment))
                        .WithTrailingTrivia(node.GetTrailingTrivia());
                }
                else if (_replacedBooleanPropertiesLookup.TryGetValue(variableName, out replacement))
                {
                    if (!replacement.Contains("("))
                    {
                        // HACK: Treat simple boolean replacements as being reversed in meaning.
                        // If any get added that are not reversed, this logic will need to change.
                        ExpressionSyntax newRight = SyntaxFactory
                            .ParseExpression($"!({node.Right.WithoutTrivia()})")
                            .WithTriviaFrom(node.Right);

                        node = node.WithRight(newRight);
                    }
                }

                return node;
            }

            public override SyntaxNode VisitPropertyDeclaration(PropertyDeclarationSyntax node)
            {
                node = (PropertyDeclarationSyntax)base.VisitPropertyDeclaration(node);

                AttributeSyntax displayNameAttribute = GetAttribute(node, "DisplayName")
                                                       ?? GetAttribute(node, "Design.DisplayName")
                                                       ?? GetAttribute(node, "Gui.Design.DisplayName");


                AttributeSyntax descriptionAttribute = GetAttribute(node, "Description");
                AttributeSyntax categoryAttribute = GetAttribute(node, "GridCategory") ?? GetAttribute(node, "Category");
                bool hasDisplayAttribute = displayNameAttribute != null || descriptionAttribute != null ||
                                           categoryAttribute != null;

                if (hasDisplayAttribute)
                {
                    node = ExcludeAttributes(node, displayNameAttribute, descriptionAttribute, categoryAttribute);
                    SyntaxList<AttributeListSyntax> newAttributes = node.AttributeLists;

                    SyntaxTriviaList leadingWhitespace = SyntaxFactory
                        .TriviaList(
                            node.GetLeadingTrivia()
                                .TakeWhile(
                                    t => t.IsKind(SyntaxKind.WhitespaceTrivia) || t.IsKind(SyntaxKind.EndOfLineTrivia)));

                    string text = "[NinjaScriptProperty]";
                    AttributeListSyntax ninjaScriptProperty = CreateAttribute(text);

                    if (node.AttributeLists.Count == 0 &&
                        node.GetLeadingTrivia()
                            .Any(t => !t.IsKind(SyntaxKind.WhitespaceTrivia) && !t.IsKind(SyntaxKind.EndOfLineTrivia)))
                    {
                        ninjaScriptProperty = ninjaScriptProperty.WithLeadingTrivia(node.GetLeadingTrivia());
                        node = node.WithLeadingTrivia(leadingWhitespace);
                    }
                    else
                        ninjaScriptProperty = ninjaScriptProperty.WithLeadingTrivia(leadingWhitespace);

                    newAttributes = newAttributes.Add(ninjaScriptProperty);

                    var parameters = new List<string>();

                    if (displayNameAttribute != null)
                        parameters.Add($"Name = {GetAttributeArgument(displayNameAttribute, 0)}");

                    if (descriptionAttribute != null)
                        parameters.Add($"Description = {GetAttributeArgument(descriptionAttribute, 0)}");

                    if (categoryAttribute != null)
                        parameters.Add($"GroupName = {GetAttributeArgument(categoryAttribute, 0)}");

                    parameters.Add("Order = 1");
                    text = string.Format("[Display({0})]", string.Join(", ", parameters));
                    AttributeListSyntax display = CreateAttribute(text).WithLeadingTrivia(leadingWhitespace);

                    if (node.AttributeLists.Count != 0)
                        display = display.WithTrailingTrivia(SyntaxFactory.CarriageReturnLineFeed);

                    newAttributes = newAttributes.Add(display);

                    node = node.WithAttributeLists(newAttributes);
                }

                return node;
            }

            private Dictionary<string, string> _constructorLookup;
            private Dictionary<string, string> _membersLookup;
            private Dictionary<string, string> _virtualMethodLookup;
            private HashSet<string> _removedMethodsLookup;
            private HashSet<string> _removedPropertiesLookup;
            private Dictionary<string, string> _replacedBooleanPropertiesLookup;
            private HashSet<string> _drawMethodsLookup;

            private string GetAttributeArgument(AttributeSyntax attribute, int argumentIndex)
            {
                return attribute?.ArgumentList.Arguments[argumentIndex].Expression.ToString();
            }

            private AttributeSyntax GetAttribute(TypeDeclarationSyntax node, string attributeName)
            {
                string fullAttributeName = attributeName + "Attribute";

                IEnumerable<AttributeSyntax> attributes =
                    from list in node.AttributeLists
                    from attribute in list.Attributes
                    let name = attribute.Name.ToString()
                    where name == attributeName || name == fullAttributeName
                    select attribute;

                return attributes.FirstOrDefault();
            }

            private AttributeSyntax GetAttribute(PropertyDeclarationSyntax node, string attributeName)
            {
                string fullAttributeName = attributeName + "Attribute";

                IEnumerable<AttributeSyntax> attributes =
                    from list in node.AttributeLists
                    from attribute in list.Attributes
                    let name = attribute.Name.ToString()
                    where name == attributeName || name == fullAttributeName
                    select attribute;

                return attributes.FirstOrDefault();
            }

            private ClassDeclarationSyntax ExcludeAttributes(ClassDeclarationSyntax node,
                params AttributeSyntax[] exclude)
            {
                var excludeLookup =
                    new HashSet<AttributeListSyntax>(
                        exclude.Where(e => e != null).Select(e => e.Parent as AttributeListSyntax));
                IEnumerable<AttributeListSyntax> included =
                    node.AttributeLists.Where(list => !excludeLookup.Contains(list));
                IEnumerable<AttributeListSyntax> excluded =
                    node.AttributeLists.Where(list => excludeLookup.Contains(list));
                SyntaxList<AttributeListSyntax> newAttributes = SyntaxFactory.List<AttributeListSyntax>(included);

                AttributeListSyntax listWithTrivia = excluded
                    .FirstOrDefault(list => list.GetLeadingTrivia().Any(t => !t.IsKind(SyntaxKind.WhitespaceTrivia)));

                if (listWithTrivia != null)
                {
                    if (newAttributes.Count != 0)
                    {
                        newAttributes = newAttributes
                            .Replace(newAttributes[0],
                                newAttributes[0].WithLeadingTrivia(listWithTrivia.GetLeadingTrivia()));

                        node = node.WithAttributeLists(newAttributes);
                    }
                    else
                    {
                        node = node
                            .WithAttributeLists(newAttributes)
                            .WithLeadingTrivia(listWithTrivia.GetLeadingTrivia());
                    }
                }
                else
                    node = node.WithAttributeLists(newAttributes);

                return node;
            }

            private PropertyDeclarationSyntax ExcludeAttributes(PropertyDeclarationSyntax node,
                params AttributeSyntax[] exclude)
            {
                var excludeLookup =
                    new HashSet<AttributeListSyntax>(
                        exclude.Where(e => e != null).Select(e => e.Parent as AttributeListSyntax));
                IEnumerable<AttributeListSyntax> included =
                    node.AttributeLists.Where(list => !excludeLookup.Contains(list));
                IEnumerable<AttributeListSyntax> excluded =
                    node.AttributeLists.Where(list => excludeLookup.Contains(list));
                SyntaxList<AttributeListSyntax> newAttributes = SyntaxFactory.List<AttributeListSyntax>(included);

                AttributeListSyntax listWithTrivia = excluded
                    .FirstOrDefault(list => list.GetLeadingTrivia().Any(t => !t.IsKind(SyntaxKind.WhitespaceTrivia)));

                if (listWithTrivia != null)
                {
                    if (newAttributes.Count != 0)
                    {
                        newAttributes = newAttributes
                            .Replace(newAttributes[0],
                                newAttributes[0].WithLeadingTrivia(listWithTrivia.GetLeadingTrivia()));

                        node = node.WithAttributeLists(newAttributes);
                    }
                    else
                    {
                        node = node
                            .WithAttributeLists(newAttributes)
                            .WithLeadingTrivia(listWithTrivia.GetLeadingTrivia());
                    }
                }
                else
                    node = node.WithAttributeLists(newAttributes);

                return node;
            }

            private SyntaxNode CreateSyntaxNode(string text)
            {
                return SyntaxFactory.ParseSyntaxTree(text).GetRoot();
            }

            private AttributeListSyntax CreateAttribute(string attributeText)
            {
                var root = CreateSyntaxNode(attributeText);

                return root
                    .DescendantNodes()
                    .First(n => n.IsKind(SyntaxKind.AttributeList)) as AttributeListSyntax;
            }

            private bool GetBooleanPropertyReplacement(string name, bool trueValue, out ExpressionSyntax replacement)
            {
                replacement = null;
                string replacementText;

                if (!_replacedBooleanPropertiesLookup.TryGetValue(name, out replacementText)) return false;

                string[] part = replacementText.Split(',');
                replacement = SyntaxFactory.ParseExpression(trueValue ? part[0] : part[1]);
                return true;
            }

            private ExpressionSyntax ApplyReplacedBooleanProperty(ExpressionSyntax node)
            {
                var urinaryExpression = node as PrefixUnaryExpressionSyntax;
                var identifierName = node as NameSyntax;
                var memberAccess = node as MemberAccessExpressionSyntax;
                SyntaxNode nameNode = ((SyntaxNode)identifierName ?? memberAccess);
                ExpressionSyntax replacement = null;

                if (nameNode != null)
                {
                    GetBooleanPropertyReplacement(nameNode.ToString(), true, out replacement);
                }
                else if ((urinaryExpression?.IsKind(SyntaxKind.LogicalNotExpression)).GetValueOrDefault())
                {
                    identifierName = urinaryExpression.Operand as NameSyntax;
                    memberAccess = urinaryExpression.Operand as MemberAccessExpressionSyntax;
                    nameNode = ((SyntaxNode)identifierName ?? memberAccess);

                    if (nameNode != null)
                        GetBooleanPropertyReplacement(nameNode.ToString(), false, out replacement);
                }

                if (replacement != null)
                    node = replacement.WithTriviaFrom(nameNode);

                return node;
            }
        }

        #region Helpers

        private static IEnumerable<T> GetDescendantNodesOfTypeWithPropertyName<T>(SyntaxNode node, params string[] names)
        {
            var namesLookup = new HashSet<string>(names);

            return node
                .DescendantNodes()
                .SelectMany(n => GetProperties(n, namesLookup))
                .OfType<T>();
        }

        private static IEnumerable<object> GetProperties(SyntaxNode node, HashSet<string> namesLookup)
        {
            return node
                .GetType()
                .GetProperties(BindingFlags.Public | BindingFlags.Instance)
                .Where(p => namesLookup.Contains(p.Name))
                .Select(p => p.GetValue(node));
        }

        private static SyntaxNode CreateSyntaxNode(string text)
        {
            return SyntaxFactory.ParseSyntaxTree(text).GetRoot();
        }

        private static T GetFirstChildNode<T>(SyntaxNode node) where T : class
        {
            return node?.ChildNodes().OfType<T>().FirstOrDefault();
        }

        private static T GetLastChildNode<T>(SyntaxNode node) where T : class
        {
            return node?.ChildNodes().OfType<T>().LastOrDefault();
        }

        private static T GetFirstDescendantNode<T>(SyntaxNode node) where T : class
        {
            return node?.DescendantNodes().OfType<T>().FirstOrDefault();
        }

        private static T GetLastDescendantNode<T>(SyntaxNode node) where T : class
        {
            return node?.DescendantNodes().OfType<T>().LastOrDefault();
        }

        #endregion
    }
}
